# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019-2020 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Optimization problem description classes.

This module implements a footprint of a single problem, storing number and types
of expressions and constraints, and the specification of a problem class.
"""

import math
from inspect import isclass
from itertools import chain

from .. import constraints, expressions, glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from ..containers import RecordTree
from ..expressions import variables
# from .options import OPTION_OBJS, Options  # Circular import.

_API_START = api_start(globals())
# -------------------------------


# Constants used by both Footprint and Specification.
_OBJS = tuple(exp for exp in expressions.__dict__.values() if isclass(exp)
    and issubclass(exp, expressions.Expression)
    and exp != expressions.Expression)  # ALl proper Expression subclasses.
_DIRS = ("find", "min", "max")
_VARS = tuple(var for var in expressions.__dict__.values() if isclass(var)
    and issubclass(var, expressions.BaseVariable)
    and var != expressions.BaseVariable)  # ALl proper BaseVariable subclasses.
_CONS = tuple(con for con in constraints.__dict__.values() if isclass(con)
    and issubclass(con, constraints.Constraint)
    and con != constraints.Constraint)  # All proper Constraint subclasses.
# _OPTS = tuple(option.name for option in OPTION_OBJS)  # All option names.


class Footprint(RecordTree):
    """Statistics on an optimization problem.

    Subtree comparison (`<<`) can be used to check if a specific problem
    represented by a :class:`Footprint` instance is contained in a problem class
    represented by a :class:`Specification` instance.
    """

    _ADD_VALUES = [("var",), ("con",)]

    def __init__(self, recordsOrDict):
        """Construct a :class:`Footprint` from raw data.

        See :class:`picos.containers.RecordTree.__init__`. The ``addValues``
        argument is fixed; only variable and constraint paths are added.
        """
        super(Footprint, self).__init__(recordsOrDict, self._ADD_VALUES)

        # Sanity check records.
        for path, value in self.items:
            pathLen  = len(path)
            category = path[0] if pathLen > 0 else None
            suptype  = path[1] if pathLen > 1 else None
            subtype  = path[2] if pathLen > 2 else None

            if category == "dir":
                if pathLen == 2 and suptype in _DIRS and value is None:
                    continue
            elif category == "obj":
                if pathLen == 3 and suptype in _OBJS \
                and isinstance(subtype, suptype.Subtype) and value is None:
                    continue
            elif category == "var":
                if pathLen == 3 and suptype in _VARS \
                and isinstance(subtype, suptype.VarSubtype) \
                and isinstance(value, int):
                    continue
            elif category == "con":
                if pathLen == 3 and suptype in _CONS \
                and isinstance(subtype, suptype.Subtype) \
                and isinstance(value, int):
                    continue
            elif category == "opt":
                # if pathLen == 2 and suptype in _OPTS:
                if pathLen == 2 and isinstance(suptype, str):
                    continue

            raise ValueError("Invalid problem footprint record: {} = {}."
                .format(path, value))

        # Sanity check for inconsistent duplicates or missing fields.
        if len(self["dir"]) != 1:
            raise TypeError("Not exactly one optimization direction defined for"
                " a problem footprint (but {}).".format(len(self["dir"])))
        elif len(self["obj"]) != 1:
            raise TypeError("Not exactly one objective function defined for a "
                "problem footprint (but {}).".format(len(self["obj"])))

    def updated(self, recordsOrDict):
        """Override :class:`~picos.containers.RecordTree.updated`.

        This method, just like :meth:`__init__`, does not take the additional
        ``addValues`` argument.
        """
        return self.__class__(
            chain(self.records, self._record_iterator(recordsOrDict)))

    @property
    def direction(self):
        """Objective function optimization direction."""
        return next(self["dir"].paths)[0]

    @property
    def objective(self):
        """Detailed type of the objective function."""
        clstype, subtype = next(self["obj"].paths)
        return expressions.ExpressionType(clstype, subtype)

    @cached_property
    def variables(self):
        """A dictionary mapping detailed variable types to their quantity."""
        return {variables.VariableType(*ts): n
            for ts, n in self.get("var").items}

    @cached_property
    def constraints(self):
        """A dictionary mapping detailed constraint types to their quantity."""
        return {constraints.ConstraintType(*ts): n
            for ts, n in self.get("con").items}

    @cached_property
    def nondefault_options(self):
        """A dictionary mapping option names to their nondefault values.

        .. warning::

            This property is cached for performance reasons, do not modify any
            mutable option value (make a deep copy instead)!
        """
        return {path[0]: value for path, value in self.get("opt").items}

    @cached_property
    def options(self):
        """An :class:`~picos.Options` object.

        .. warning::

            This property is cached for performance reasons, do not modify the
            returned object (make a :meth:`~picos.Options.copy` instead)!
        """
        from .options import Options
        return Options(**self.nondefault_options)

    @cached_property
    def integer(self):
        """Whether an integral variable type is present."""
        return any(("var", vtype) in self for vtype in (
            expressions.BinaryVariable, expressions.IntegerVariable))

    @property
    def continuous(self):
        """Whether no integral variable type is present."""
        return not self.integer

    @cached_property
    def nonconvex_quadratic_objective(self):
        """Whether the problem has a nonconvex quadratic objective."""
        if self.objective.clstype is expressions.QuadraticExpression:
            direction = self.direction
            subtype   = self.objective.subtype

            assert direction in ("min", "max")

            if   direction == "min" and not subtype.convex:
                return True
            elif direction == "max" and not subtype.concave:
                return True

        return False

    def __str__(self):
        dirStr = self.direction.capitalize()
        objStr = str(self.objective)
        varStr = ", ".join(sorted("{} {}".format(n, v)
            for v, n in self.variables.items()))
        conStr = ", ".join(sorted("{} {}".format(n, c)
            for c, n in self.constraints.items()))
        optStr = ", ".join(sorted("{}={}".format(n, v)
            for n, v in self.nondefault_options.items()))

        return "{} {} subject to {} using {} and {}.".format(
            dirStr, objStr,
            conStr if conStr else "no constraints",
            varStr if varStr else "no variables",
            optStr if optStr else "default options")

    def __repr__(self):
        return glyphs.repr2("Footprint", str(self))

    @classmethod
    def from_problem(cls, problem):
        """Create a footprint from a problem instance."""
        return cls(chain(
            (("dir", problem.objective.direction, None),
             ("obj", problem.objective.normalized.function.type, None)),
            (("var", v.var_type, 1) for v in problem.variables.values()),
            (("con", con.type, 1) for con in problem.constraints.values()),
            (("opt", n, v) for n, v in problem.options.nondefaults.items())))

    @classmethod
    def from_types(cls, obj_dir, obj_func, vars=[], cons=[], nd_opts={}):
        """Create a footprint from collections of detailed types.

        :param str obj_dir:
            Objective direction.

        :param obj_func:
            Detailed objective function type.

        :parm list(tuple) vars:
            A list of pairs of detailed variable type and occurence count.

        :parm list(tuple) cons:
            A list of pairs of detailed constraint type and occurence count.

        :param list(str) nd_opts:
            A dictionary mapping option names to nondefault values.
        """
        return cls(chain(
            (("dir", obj_dir, None),
             ("obj", obj_func, None)),
            (("var", vn[0], vn[1]) for vn in vars),
            (("con", cn[0], cn[1]) for cn in cons),
            (("opt", name, value) for name, value in nd_opts.items())))

    def with_extra_options(self, **extra_options):
        """Return a copy with additional solution search options applied."""
        # Determine the new option set.
        options = self.options.self_or_updated(**extra_options)

        # Delete old nondefault options and set new ones.
        return self.updated(chain(
            (("opt", self.NONE),),
            (("opt", n, v) for n, v in options.nondefaults.items())
        ))

    @cached_property
    def size(self):
        """Return the estimated size of the (dense) scalar constraint matrix."""
        num_vars = 0
        num_cons = 0

        for dim in self.variables.values():
            num_vars += dim

        for con, num in self.constraints.items():
            num_cons += con.clstype._cost(con.subtype)*num

        return max(1, num_vars)*max(1, num_cons)

    @property
    def cost(self):
        """A very rough measure on how expensive solving such a problem is.

        This is logarithmic in the estimated size of the constraint matrix.
        """
        return math.log(self.size, 10)


class Specification(RecordTree):
    """Representation of a mathematical class of optimization problems.

    Subtree comparison (`<<`) can be used to check if a specific problem
    represented by a :class:`Footprint` instance is contained in a problem class
    represented by a :class:`Specification` instance.
    """

    def __init__(self, recordsOrDict):
        """Construct a :class:`Specification` from raw data.

        See :class:`picos.containers.RecordTree.__init__`.
        """
        super(Specification, self).__init__(recordsOrDict)

        # Sanity check records (up to expression and constraint subtypes).
        for path, value in self.items:
            pathLen  = len(path)
            category = path[0] if pathLen > 0 else None
            suptype  = path[1] if pathLen > 1 else None
            subtype  = path[2] if pathLen > 2 else None

            if category and pathLen == 1 and value is self.ALL:
                continue

            if category == "dir":
                if pathLen == 2 and suptype in _DIRS and value is None:
                    continue
            elif category == "obj":
                if pathLen == 3 and suptype in _OBJS \
                and isinstance(subtype, suptype.Subtype) and value is None:
                    continue
                elif pathLen == 2 and suptype in _OBJS and value is self.ALL:
                    continue
            elif category == "var":
                if pathLen == 3 and suptype in _VARS \
                and isinstance(subtype, suptype.VarSubtype) and value is None:
                    continue
                elif pathLen == 2 and suptype in _VARS and value is self.ALL:
                    continue
            elif category == "con":
                if pathLen == 3 and suptype in _CONS \
                and isinstance(subtype, suptype.Subtype) and value is None:
                    continue
                elif pathLen == 2 and suptype in _CONS and value is self.ALL:
                    continue
            elif category == "opt":
                # if pathLen == 3 and suptype in _OPTS and value is None:
                if pathLen == 3 and isinstance(suptype, str) and value is None:
                    continue
                # if pathLen == 2 and suptype in _OPTS and value is self.ALL:
                if pathLen == 2 and isinstance(suptype, str) \
                and value is self.ALL:
                    continue

            raise ValueError("Invalid problem specification record: {} = {}."
                .format(path, value))

    @staticmethod
    def _paths_str(paths):
        return ", ".join(":".join(p.__name__ if isclass(p) else str(p) for p in
            path) for path in paths)

    def __str__(self):
        dirs, objs, vars, cons, opts = \
            (self.get(x) for x in ("dir", "obj", "var", "con", "opt"))

        if not dirs:
            dirStr = "Load"
        elif dirs is self.ALL:
            dirStr = "Optimize"
        else:
            dirStr = self._paths_str(dirs.paths).capitalize()

        if not objs:
            objStr = "no objective"
        elif objs is self.ALL:
            objStr = "any objective"
        else:
            objStr = self._paths_str(objs.paths)

        if not vars:
            varStr = "no variables"
        elif vars is self.ALL:
            varStr = "any variables"
        else:
            varStr = self._paths_str(vars.paths)

        if not cons:
            conStr = "no constraint"
        elif cons is self.ALL:
            conStr = "any constraint"
        else:
            conStr = self._paths_str(cons.paths)

        if not opts:
            optStr = "default options"
        elif opts is self.ALL:
            optStr = "any options"
        else:
            optStr = "nondefault values for " + self._paths_str(opts.paths)

        return "{} {} subject to {} using {} and {}.".format(
            dirStr, objStr, conStr, varStr, optStr)

    def __repr__(self):
        return glyphs.repr2("Specification", str(self))

    @classmethod
    def compile(cls, directions=(), objectives=(), variables=(),
            constraints=(), nondefault_options=RecordTree.ALL):
        """Create a specification from the given features."""
        return cls(chain(
            (("dir", cls.ALL),) if directions is cls.ALL else
            (("dir", dir, None) for dir in directions),

            (("obj", cls.ALL),) if objectives is cls.ALL else
            (("obj", obj, cls.ALL) for obj in objectives),

            (("var", cls.ALL),) if variables is cls.ALL else
            (("var", var, cls.ALL) for var in variables),

            (("con", cls.ALL),) if constraints is cls.ALL else
            (("con", con, cls.ALL) for con in constraints),

            (("opt", cls.ALL),) if nondefault_options is cls.ALL else
            (("opt", opt, cls.ALL) for opt in nondefault_options)
        ))


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C)      2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Functions for reading optimization problems from a file."""

# ------------------------------------------------------------------------------
# TODO: This file is using ad-hoc parser implementations. Use a proper parser
#       generation library and a unified 'read' function for different formats.
# ------------------------------------------------------------------------------

import warnings

import cvxopt
import numpy

from ..apidoc import api_end, api_start
from ..expressions import AffineExpression, Constant
from .problem import Problem

_API_START = api_start(globals())
# -------------------------------


def _spmatrix(*args, **kwargs):
    """Create a CVXOPT sparse matrix.

    A wrapper around :func:`cvxopt.spmatrix` that converts indices to
    :class:`int`, if necessary.

    Works around PICOS sometimes passing indices as :class:`numpy:numpy.int64`.
    """
    try:
        return cvxopt.spmatrix(*args, **kwargs)
    except TypeError as error:
        # CVXOPT does not like NumPy's int64 scalar type for indices, so attempt
        # to convert all indices to Python's int.
        if str(error) == "non-numeric type in list":
            newargs = list(args)

            for argNum, arg in enumerate(args):
                if argNum in (1, 2):  # Positional I, J.
                    newargs[argNum] = [int(x) for x in args[argNum]]

            for kw in "IJ":
                if kw in kwargs:
                    kwargs[kw] = [int(x) for x in kwargs[kw]]

            return cvxopt.spmatrix(*newargs, **kwargs)
        else:
            raise


def _break_cols(mat, sizes):
    """Help with the reading of CBF files."""
    n = len(sizes)
    I, J, V = [], [], []
    for i in range(n):
        I.append([])
        J.append([])
        V.append([])
    cumsz = numpy.cumsum(sizes)
    import bisect

    for i, j, v in zip(mat.I, mat.J, mat.V):
        block = bisect.bisect(cumsz, j)
        I[block].append(i)
        V[block].append(v)
        if block == 0:
            J[block].append(j)
        else:
            J[block].append(j - cumsz[block - 1])
    return [
        _spmatrix(V[k], I[k], J[k], (mat.size[0], sz))
        for k, sz in enumerate(sizes)
    ]


def _break_rows(mat, sizes):
    """Help with the reading of CBF files."""
    n = len(sizes)
    I, J, V = [], [], []
    for i in range(n):
        I.append([])
        J.append([])
        V.append([])
    cumsz = numpy.cumsum(sizes)
    import bisect

    for i, j, v in zip(mat.I, mat.J, mat.V):
        block = bisect.bisect(cumsz, i)
        J[block].append(j)
        V[block].append(v)
        if block == 0:
            I[block].append(i)
        else:
            I[block].append(i - cumsz[block - 1])
    return [
        _spmatrix(V[k], I[k], J[k], (sz, mat.size[1]))
        for k, sz in enumerate(sizes)
    ]


def _block_idx(i, sizes):
    """Help with the reading of CBF files."""
    # if there are blocks of sizes n1,...,nk and i is
    # the index of an element of the big vectorized variable,
    # returns the block of i and its index inside the sub-block.
    cumsz = numpy.cumsum(sizes)
    import bisect

    block = bisect.bisect(cumsz, i)
    return block, (i if block == 0 else i - cumsz[block - 1])


def _read_cbf_block(P, blocname, f, parsed_blocks):
    """Help with the reading of CBF files."""
    if blocname == "OBJSENSE":
        objsense = f.readline().split()[0].lower()
        P.set_objective(objsense, P.objective.normalized.function)
        return None
    elif blocname == "PSDVAR":
        n = int(f.readline())
        vardims = []
        XX = []
        for i in range(n):
            ni = int(f.readline())
            vardims.append(ni)
            Xi = P.add_variable("X[" + str(i) + "]", (ni, ni), "symmetric")
            XX.append(Xi)
            P.add_constraint(Xi >> 0)
        return vardims, XX
    elif blocname == "VAR":
        Nscalar, ncones = [int(fi) for fi in f.readline().split()]
        tot_dim = 0
        var_structure = []
        xx = []
        for i in range(ncones):
            lsplit = f.readline().split()
            tp, dim = lsplit[0], int(lsplit[1])
            tot_dim += dim
            var_structure.append(dim)
            if tp == "F":
                xi = P.add_variable("x[" + str(i) + "]", dim)
            elif tp == "L+":
                xi = P.add_variable("x[" + str(i) + "]", dim, lower=0)
            elif tp == "L-":
                xi = P.add_variable("x[" + str(i) + "]", dim, upper=0)
            elif tp == "L=":
                xi = P.add_variable("x[" + str(i) + "]", dim, lower=0, upper=0)
            elif tp == "Q":
                xi = P.add_variable("x[" + str(i) + "]", dim)
                P.add_constraint(abs(xi[1:]) < xi[0])
            elif tp == "QR":
                xi = P.add_variable("x[" + str(i) + "]", dim)
                P.add_constraint(abs(xi[2:]) ** 2 < 2 * xi[0] * xi[1])
            xx.append(xi)
        if tot_dim != Nscalar:
            raise Exception("VAR dimensions do not match the header")
        return Nscalar, var_structure, xx
    elif blocname == "INT":
        n = int(f.readline())
        ints = {}
        for k in range(n):
            j = int(f.readline())
            i, col = _block_idx(j, parsed_blocks["VAR"][1])
            ints.setdefault(i, [])
            ints[i].append(col)
        x = parsed_blocks["VAR"][2]
        for i in ints:
            if len(ints[i]) == x[i].size[0]:
                x[i].vtype = "integer"
            else:
                x.append(
                    P.add_variable(
                        "x_int[" + str(i) + "]", len(ints[i]), "integer"
                    )
                )
                for k, j in enumerate(ints[i]):
                    P.add_constraint(x[i][j] == x[-1][k])
        return x
    elif blocname == "CON":
        Ncons, ncones = [int(fi) for fi in f.readline().split()]
        cons_structure = []
        tot_dim = 0
        for i in range(ncones):
            lsplit = f.readline().split()
            tp, dim = lsplit[0], int(lsplit[1])
            tot_dim += dim
            cons_structure.append((tp, dim))
        if tot_dim != Ncons:
            raise Exception("CON dimensions do not match the header")
        return Ncons, cons_structure
    elif blocname == "PSDCON":
        n = int(f.readline())
        psdcons_structure = []
        for i in range(n):
            ni = int(f.readline())
            psdcons_structure.append(ni)
        return psdcons_structure
    elif blocname == "OBJACOORD":
        n = int(f.readline())
        J = []
        V = []
        for i in range(n):
            lsplit = f.readline().split()
            j, v = int(lsplit[0]), float(lsplit[1])
            J.append(j)
            V.append(v)
        return _spmatrix(V, [0] * len(J), J, (1, parsed_blocks["VAR"][0]))
    elif blocname == "OBJBCOORD":
        return float(f.readline())
    elif blocname == "OBJFCOORD":
        n = int(f.readline())
        Fobj = [
            _spmatrix([], [], [], (ni, ni)) for ni in parsed_blocks["PSDVAR"][0]
        ]
        for k in range(n):
            lsplit = f.readline().split()
            j, row, col, v = (
                int(lsplit[0]),
                int(lsplit[1]),
                int(lsplit[2]),
                float(lsplit[3]),
            )
            Fobj[j][row, col] = v
            if row != col:
                Fobj[j][col, row] = v
        return Fobj
    elif blocname == "FCOORD":
        n = int(f.readline())
        Fblocks = {}
        for k in range(n):
            lsplit = f.readline().split()
            i, j, row, col, v = (
                int(lsplit[0]),
                int(lsplit[1]),
                int(lsplit[2]),
                int(lsplit[3]),
                float(lsplit[4]),
            )
            if i not in Fblocks:
                Fblocks[i] = [
                    _spmatrix([], [], [], (ni, ni))
                    for ni in parsed_blocks["PSDVAR"][0]
                ]
            Fblocks[i][j][row, col] = v
            if row != col:
                Fblocks[i][j][col, row] = v
        return Fblocks
    elif blocname == "ACOORD":
        n = int(f.readline())
        J = []
        V = []
        I = []
        for k in range(n):
            lsplit = f.readline().split()
            i, j, v = int(lsplit[0]), int(lsplit[1]), float(lsplit[2])
            I.append(i)
            J.append(j)
            V.append(v)
        return _spmatrix(
            V, I, J, (parsed_blocks["CON"][0], parsed_blocks["VAR"][0])
        )
    elif blocname == "BCOORD":
        n = int(f.readline())
        V = []
        I = []
        for k in range(n):
            lsplit = f.readline().split()
            i, v = int(lsplit[0]), float(lsplit[1])
            I.append(i)
            V.append(v)
        return _spmatrix(V, I, [0] * len(I), (parsed_blocks["CON"][0], 1))
    elif blocname == "HCOORD":
        n = int(f.readline())
        Hblocks = {}
        for k in range(n):
            lsplit = f.readline().split()
            i, j, row, col, v = (
                int(lsplit[0]),
                int(lsplit[1]),
                int(lsplit[2]),
                int(lsplit[3]),
                float(lsplit[4]),
            )
            if j not in Hblocks:
                Hblocks[j] = [
                    _spmatrix([], [], [], (ni, ni))
                    for ni in parsed_blocks["PSDCON"]
                ]
            Hblocks[j][i][row, col] = v
            if row != col:
                Hblocks[j][i][col, row] = v
        return Hblocks
    elif blocname == "DCOORD":
        n = int(f.readline())
        Dblocks = [
            _spmatrix([], [], [], (ni, ni)) for ni in parsed_blocks["PSDCON"]
        ]
        for k in range(n):
            lsplit = f.readline().split()
            i, row, col, v = (
                int(lsplit[0]),
                int(lsplit[1]),
                int(lsplit[2]),
                float(lsplit[3]),
            )
            Dblocks[i][row, col] = v
            if row != col:
                Dblocks[i][col, row] = v
        return Dblocks
    else:
        raise Exception("unexpected block name")


def import_cbf(filename):
    """Create a :class:`~picos.Problem` from a CBF file.

    The created problem contains one (multidimensional) variable for each cone
    specified in the section ``VAR`` of the .cbf file, and one
    (multidimmensional) constraint for each cone specified in the sections
    ``CON`` and ``PSDCON``.

    :returns: A tuple ``(P, x, X, params)`` where

        - ``P`` is the imported picos :class:`~picos.Problem` object,
        - ``x`` is a list of multidimensional variables representing the scalar
          variables found in the file,
        - ``X`` is a list of symmetric variables representing the positive
          semidefinite variables found in the file, and
        - ``params`` is a dictionary containing PICOS cosntants used to define
          the problem. Indexing is with respect to the blocks of variables as
          defined in the sections ``VAR`` and  ``CON`` of the file.
    """
    P = Problem()

    try:
        f = open(filename, "r")
    except IOError:
        filename += ".cbf"
        f = open(filename, "r")

    line = f.readline()
    while not line.startswith("VER"):
        line = f.readline()

    ver = int(f.readline())
    if ver != 1:
        warnings.warn("CBF file has a version other than 1.")

    structure_keywords = ["OBJSENSE", "PSDVAR", "VAR", "INT", "PSDCON", "CON"]
    data_keywords = [
        "OBJFCOORD",
        "OBJACOORD",
        "OBJBCOORD",
        "FCOORD",
        "ACOORD",
        "BCOORD",
        "HCOORD",
        "DCOORD",
    ]

    structure_mode = True  # still parsing structure blocks
    seen_blocks = []
    parsed_blocks = {}
    while True:
        line = f.readline()
        if not line:
            break
        lsplit = line.split()
        if lsplit and lsplit[0] in structure_keywords:
            if lsplit[0] == "INT" and ("VAR" not in seen_blocks):
                raise Exception("INT BLOCK before VAR BLOCK")
            if lsplit[0] == "CON" and not (
                "VAR" in seen_blocks or "PSDVAR" in seen_blocks
            ):
                raise Exception("CON BLOCK before VAR/PSDVAR BLOCK")
            if lsplit[0] == "PSDCON" and not (
                "VAR" in seen_blocks or "PSDVAR" in seen_blocks
            ):
                raise Exception("PSDCON BLOCK before VAR/PSDVAR BLOCK")
            if lsplit[0] == "VAR" and (
                "CON" in seen_blocks or "PSDCON" in seen_blocks
            ):
                raise Exception("VAR BLOCK after CON/PSDCON BLOCK")
            if lsplit[0] == "PSDVAR" and (
                "CON" in seen_blocks or "PSDCON" in seen_blocks
            ):
                raise Exception("PSDVAR BLOCK after CON/PSDCON BLOCK")
            if structure_mode:
                parsed_blocks[lsplit[0]] = _read_cbf_block(
                    P, lsplit[0], f, parsed_blocks
                )
                seen_blocks.append(lsplit[0])
            else:
                raise Exception("Structure keyword after first data item")
        if lsplit and lsplit[0] in data_keywords:
            if "OBJSENSE" not in seen_blocks:
                raise Exception("missing OBJSENSE block")
            if not ("VAR" in seen_blocks or "PSDVAR" in seen_blocks):
                raise Exception("missing VAR/PSDVAR block")
            if lsplit[0] in ("OBJFCOORD", "FCOORD") and not (
                "PSDVAR" in seen_blocks
            ):
                raise Exception("missing PSDVAR block")
            if lsplit[0] in ("OBJACOORD", "ACOORD", "HCOORD") and not (
                "VAR" in seen_blocks
            ):
                raise Exception("missing VAR block")
            if lsplit[0] in ("DCOORD", "HCOORD") and not (
                "PSDCON" in seen_blocks
            ):
                raise Exception("missing PSDCON block")
            structure_mode = False
            parsed_blocks[lsplit[0]] = _read_cbf_block(
                P, lsplit[0], f, parsed_blocks
            )
            seen_blocks.append(lsplit[0])

    f.close()
    # variables
    if "VAR" in parsed_blocks:
        Nvars, varsz, x = parsed_blocks["VAR"]
    else:
        x = None

    if "INT" in parsed_blocks:
        x = parsed_blocks["INT"]

    if "PSDVAR" in parsed_blocks:
        psdsz, X = parsed_blocks["PSDVAR"]
    else:
        X = None

    # objective
    obj_constant = parsed_blocks.get("OBJBCOORD", 0)
    bobj = Constant("bobj", obj_constant)
    obj = Constant("bobj", obj_constant)

    aobj = {}
    if "OBJACOORD" in parsed_blocks:
        obj_vecs = _break_cols(parsed_blocks["OBJACOORD"], varsz)
        aobj = {}
        for k, v in enumerate(obj_vecs):
            if v:
                aobj[k] = Constant("c[" + str(k) + "]", v)
                obj += aobj[k] * x[k]

    Fobj = {}
    if "OBJFCOORD" in parsed_blocks:
        Fbl = parsed_blocks["OBJFCOORD"]
        for i, Fi in enumerate(Fbl):
            if Fi:
                Fobj[i] = Constant("F[" + str(i) + "]", Fi)
                obj += Fobj[i] | X[i]

    P.set_objective(P.objective.normalized.direction, obj)

    # cone constraints
    bb = {}
    AA = {}
    FF = {}
    if "CON" in parsed_blocks:
        Ncons, structcons = parsed_blocks["CON"]
        szcons = [s for tp, s in structcons]

        b = parsed_blocks.get("BCOORD", _spmatrix([], [], [], (Ncons, 1)))
        bvecs = _break_rows(b, szcons)
        consexp = []
        for i, bi in enumerate(bvecs):
            bb[i] = Constant("b[" + str(i) + "]", bi)
            consexp.append(Constant("b[" + str(i) + "]", bi))

        A = parsed_blocks.get("ACOORD", _spmatrix([], [], [], (Ncons, Nvars)))
        Ablc = _break_rows(A, szcons)
        for i, Ai in enumerate(Ablc):
            Aiblocs = _break_cols(Ai, varsz)
            for j, Aij in enumerate(Aiblocs):
                if Aij:
                    AA[i, j] = Constant("A[" + str((i, j)) + "]", Aij)
                    consexp[i] += AA[i, j] * x[j]

        Fcoords = parsed_blocks.get("FCOORD", {})
        for k, mats in Fcoords.items():
            i, row = _block_idx(k, szcons)
            row_exp = AffineExpression.zero()
            for j, mat in enumerate(mats):
                if mat:
                    FF[i, j, row] = Constant("F[" + str((i, j, row)) + "]", mat)
                    row_exp += FF[i, j, row] | X[j]

            consexp[i] += (
                "e_" + str(row) + "(" + str(szcons[i]) + ",1)"
            ) * row_exp

        for i, (tp, sz) in enumerate(structcons):
            if tp == "F":
                continue
            elif tp == "L-":
                P.add_constraint(consexp[i] < 0)
            elif tp == "L+":
                P.add_constraint(consexp[i] > 0)
            elif tp == "L=":
                P.add_constraint(consexp[i] == 0)
            elif tp == "Q":
                P.add_constraint(abs(consexp[i][1:]) < consexp[i][0])
            elif tp == "QR":
                P.add_constraint(
                    abs(consexp[i][2:]) ** 2 < 2 * consexp[i][0] * consexp[i][1]
                )
            else:
                raise Exception("unexpected cone type")

    DD = {}
    HH = {}
    if "PSDCON" in parsed_blocks:
        Dblocks = parsed_blocks.get(
            "DCOORD",
            [_spmatrix([], [], [], (ni, ni)) for ni in parsed_blocks["PSDCON"]],
        )
        Hblocks = parsed_blocks.get("HCOORD", {})

        consexp = []
        for i, Di in enumerate(Dblocks):
            DD[i] = Constant("D[" + str(i) + "]", Di)
            consexp.append(Constant("D[" + str(i) + "]", Di))

        for j, Hj in Hblocks.items():
            i, col = _block_idx(j, varsz)
            for k, Hij in enumerate(Hj):
                if Hij:
                    HH[k, i, col] = Constant("H[" + str((k, i, col)) + "]", Hij)
                    consexp[k] += HH[k, i, col] * x[i][col]

        for exp in consexp:
            P.add_constraint(exp >> 0)

    params = {
        "aobj": aobj,
        "bobj": bobj,
        "Fobj": Fobj,
        "A": AA,
        "b": bb,
        "F": FF,
        "D": DD,
        "H": HH,
    }

    return P, x, X, params


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017, 2020 Guillaume Sagnol
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of matrix norm constraints."""

from collections import namedtuple

import cvxopt as cvx

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class MatrixNormConstraint(Constraint):
    """Upper bound on a matrix :math:`(p,q)`-norm."""

    class VectorNormConversion(ConstraintConversion):
        """Upper bound on a :math:`(p,q)`-norm constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import AffineExpression, RealVariable, Norm

            shape, pNum, pDen, qNum, qDen = subtype
            rows, cols = shape

            # HACK: Whether the bound is constant is irrelevant.
            bound = AffineExpression.make_type((1, 1), None, None)
            pNorm = Norm.make_type((rows, 1), pNum, pDen, pNum, pDen)
            qNorm = Norm.make_type((cols, 1), qNum, qDen, qNum, qDen)

            yield ("var", RealVariable.make_var_type(dim=cols, bnd=0), 1)
            yield ("con", pNorm <= bound, cols)
            yield ("con", qNorm <= bound, 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import Norm, RealVariable
            from ..modeling import Problem

            norm = con.norm
            p, q = norm.p, norm.q
            cols = norm.x.size[1]

            P = Problem()

            u = RealVariable("__u", cols)

            # Bound the p-norm of every column.
            for j in range(cols):
                P.add_constraint(Norm(norm.x[:, j], p) <= u[j])

            # Bound the q-norm of the column norms.
            P.add_constraint(Norm(u, q) <= con.upperBound)

            return P

    def __init__(self, norm, upperBound):
        """Construct a :class:`MatrixNormConstraint`.

        :param ~picos.expressions.Norm norm:
            The norm.
        :param ~picos.expressions.AffineExpression upperBound:
            The scalar upper bound.
        """
        from ..expressions import AffineExpression, Norm

        assert isinstance(norm, Norm)
        assert isinstance(upperBound, AffineExpression)
        assert len(upperBound) == 1

        assert norm.qnum is not None or norm.pnum != norm.qnum, \
            "Won't create a (p,q)-norm constraint for a p-norm."

        self.norm       = norm
        self.upperBound = upperBound

        super(MatrixNormConstraint, self).__init__(norm._typeStr)

    Subtype = namedtuple("Subtype", ("shape", "pNum", "pDen", "qNum", "qDen"))

    @classmethod
    def _cost(cls, subtype):
        return subtype.shape[0] * subtype.shape[1] + 1

    def _subtype(self):
        return self.Subtype(self.norm.x.size, self.norm.pnum, self.norm.pden,
            self.norm.qnum, self.norm.qden)

    def _expression_names(self):
        yield "norm"
        yield "upperBound"

    def _str(self):
        return glyphs.le(self.norm.string, self.upperBound.string)

    def _get_slack(self):
        return self.upperBound.value - self.norm.value


class SpectralNormConstraint(Constraint):
    """Spectral norm of a matrix."""

    class Conversion(ConstraintConversion):
        """Spectral norm constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from . import (ComplexLMIConstraint, LMIConstraint)
            m, n = subtype.shape

            if subtype.hermitian:
                if subtype.complex:
                    yield ("con", ComplexLMIConstraint.make_type(diag=n), 2)
                else:
                    yield ("con", LMIConstraint.make_type(diag=n), 2)
            else:
                if subtype.complex:
                    yield ("con", ComplexLMIConstraint.make_type(diag=n+m), 1)
                else:
                    yield ("con", LMIConstraint.make_type(diag=n+m), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..modeling import Problem
            from ..expressions import Constant

            x = con.norm.x
            m, n = x.shape
            t = con.upperBound

            In = Constant('I', cvx.spdiag([1.] * n))
            P = Problem()
            if x.hermitian:
                P.add_constraint(x << t*In)
                P.add_constraint(x >> -t * In)
            else:
                Im = Constant('I', cvx.spdiag([1.] * m))
                P.add_constraint((((t*Im) & x) // (x.H & (t*In))) >> 0)
            return P

    def __init__(self, norm, upperBound):
        """Construct a :class:`SpectralNormConstraint`.

        :param ~picos.expressions.SpectralNorm norm:
            Constrained spectral norm
        :param ~picos.expressions.AffineExpression upperBound:
            Upper bound on the expression.
        """
        from ..expressions import AffineExpression, SpectralNorm

        assert isinstance(norm, SpectralNorm)
        assert isinstance(upperBound, AffineExpression)
        assert len(upperBound) == 1

        self.norm       = norm
        self.upperBound = upperBound

        super(SpectralNormConstraint, self).__init__(norm._typeStr)

    Subtype = namedtuple("Subtype", ("shape", "complex", "hermitian"))

    def _subtype(self):
        x = self.norm.x
        return self.Subtype(x.shape, x.complex, x.hermitian)

    @classmethod
    def _cost(cls, subtype):
        m = subtype.shape[0]
        n = subtype.shape[1]

        if subtype.hermitian:
            if subtype.complex:
                return 2 * n**2 + 1
            else:
                return 2 * n*(n+1) // 2 + 1
        else:
            if subtype.complex:
                return (n + m) ** 2 + 1
            else:
                return (n + m) * (n + m + 1) // 2 + 1

    def _expression_names(self):
        yield "norm"
        yield "upperBound"

    def _str(self):
        return glyphs.le(self.norm.string, self.upperBound.string)

    def _get_slack(self):
        return self.upperBound.value - self.norm.value


class NuclearNormConstraint(Constraint):
    """Nuclear norm of a matrix."""

    class Conversion(ConstraintConversion):
        """Nuclear norm constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import SymmetricVariable, HermitianVariable
            from . import (ComplexLMIConstraint, LMIConstraint,
                           AffineConstraint, ComplexAffineConstraint)

            m, n = subtype.shape

            if subtype.complex:
                MatrixVariable = HermitianVariable
                dm, dn = m**2, n**2
            else:
                MatrixVariable = SymmetricVariable
                dm = (m * (m + 1)) // 2
                dn = (n * (n + 1)) // 2

            yield ("var", MatrixVariable.make_var_type(dim=dm, bnd=0), 1)
            yield ("var", MatrixVariable.make_var_type(dim=dn, bnd=0), 1)

            if subtype.hermitian:
                l = (n * (n + 1)) // 2  # Number of lower triangular elements.
                if subtype.complex:
                    yield ("con", ComplexAffineConstraint.make_type(dim=l), 1)
                    yield ("con", ComplexLMIConstraint.make_type(diag=n), 2)
                else:
                    yield ("con", AffineConstraint.make_type(dim=l, eq=True), 1)
                    yield ("con", LMIConstraint.make_type(diag=n), 2)
            else:
                if subtype.complex:
                    yield ("con", ComplexLMIConstraint.make_type(diag=n+m), 1)
                else:
                    yield ("con", LMIConstraint.make_type(diag=n+m), 1)

            yield ("con", AffineConstraint.make_type(dim=1, eq=False), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..modeling import Problem
            from ..expressions import SymmetricVariable, HermitianVariable
            from ..expressions.algebra import trace

            x = con.norm.x
            m, n = x.shape
            t = con.upperBound

            if con.norm._complex:
                Y = HermitianVariable("Y", (m, m))
                Z = HermitianVariable("Z", (n, n))
            else:
                Y = SymmetricVariable("Y", (m, m))
                Z = SymmetricVariable("Z", (n, n))

            P = Problem()
            if x.hermitian:
                P.add_constraint(x.trilvec == (Y - Z).trilvec)
                P.add_constraint(Y >> 0)
                P.add_constraint(Z >> 0)
                P.add_constraint(trace(Y) + trace(Z) <= t)
            else:
                P.add_constraint((Y & x) // (x.H & Z) >> 0)
                P.add_constraint(trace(Y) + trace(Z) <= 2 * t)
            return P

    def __init__(self, norm, upperBound):
        """Construct a :class:`NuclearNormConstraint`.

        :param ~picos.expressions.NuclearNorm norm:
            Constrained nuclear norm
        :param ~picos.expressions.AffineExpression upperBound:
            Upper bound on the expression.
        """
        from ..expressions import AffineExpression, NuclearNorm

        assert isinstance(norm, NuclearNorm)
        assert isinstance(upperBound, AffineExpression)
        assert len(upperBound) == 1

        self.norm       = norm
        self.upperBound = upperBound

        super(NuclearNormConstraint, self).__init__(norm._typeStr)

    Subtype = namedtuple("Subtype", ("shape", "complex", "hermitian"))

    def _subtype(self):
        x = self.norm.x
        return self.Subtype(x.shape, x.complex, x.hermitian)

    @classmethod
    def _cost(cls, subtype):
        m, n = subtype.shape

        if subtype.hermitian:
            if subtype.complex:
                return 2 * n ** 2 + 1
            else:
                return 2 * n * (n+1) // 2 + 1
        else:
            if subtype.complex:
                return (n + m) ** 2 + 1
            else:
                return (n + m) * (n + m + 1) // 2 + 1

    def _expression_names(self):
        yield "norm"
        yield "upperBound"

    def _str(self):
        return glyphs.le(self.norm.string, self.upperBound.string)

    def _get_slack(self):
        return self.upperBound.value - self.norm.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`MOSEKSolver`."""

import math
import sys

import cvxopt

from ..apidoc import api_end, api_start
from ..compat import assert_import_exists
from ..constraints import (AffineConstraint, ConvexQuadraticConstraint,
                           LMIConstraint, RSOCConstraint, SOCConstraint)
from ..expressions import (AffineExpression, BinaryVariable, IntegerVariable,
                           QuadraticExpression)
from ..modeling.footprint import Specification
from ..modeling.solution import (PS_FEASIBLE, PS_ILLPOSED, PS_INFEASIBLE,
                                 PS_UNBOUNDED, PS_UNKNOWN, SS_FEASIBLE,
                                 SS_INFEASIBLE, SS_OPTIMAL, SS_UNKNOWN)
from .solver import ProblemUpdateError, Solver

_API_START = api_start(globals())
# -------------------------------


class MOSEKSolver(Solver):
    """Interface to the MOSEK solver via its low level Optimizer API.

    Supports both MOSEK 8 and 9.

    The low level API is tedious to interface, but is currently much faster than
    the high level Fusion API, which would be the prefered interface otherwise.
    """

    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression,
            QuadraticExpression],
        variables=Specification.ALL,
        constraints=[
            AffineConstraint,
            SOCConstraint,
            RSOCConstraint,
            ConvexQuadraticConstraint,
            LMIConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.solver.Solver.supports`."""
        if not Solver.supports(footprint):
            return False

        # No nonconvex quadratic objectives.
        if footprint.nonconvex_quadratic_objective:
            return False

        conic = any(("con", constraint) in footprint
            for constraint in (SOCConstraint, RSOCConstraint, LMIConstraint))

        quadratic = footprint.objective.clstype is QuadraticExpression \
            or ("con", ConvexQuadraticConstraint) in footprint

        # No mixing of quadratic and conic problems.
        if conic and quadratic:
            return False

        # No integer SDPs.
        if footprint.integer and ("con", LMIConstraint) in footprint:
            return False

        return footprint << cls.SUPPORTED

    @classmethod
    def default_penalty(cls):
        """Implement :meth:`~.solver.Solver.default_penalty`."""
        return 0.0  # Commercial solver.

    @classmethod
    def test_availability(cls):
        """Implement :meth:`~.solver.Solver.test_availability`."""
        assert_import_exists("mosek")

    @classmethod
    def names(cls):
        """Implement :meth:`~.solver.Solver.names`."""
        return "mosek", "MOSEK (Optimizer)", "MOSEK via Optimizer API"

    def __init__(self, problem):
        """Initialize a MOSEK (Optimizer) solver interface.

        :param ~picos.Problem problem: The problem to be solved.
        """
        super(MOSEKSolver, self).__init__(problem)

        self._mosekVarOffset = dict()
        """Maps a PICOS (multidimensional) variable to a MOSEK scalar variable
        (start) index."""

        self._mosekLinConOffset = dict()
        """Maps a PICOS (multidimensional) linear constraint to a MOSEK scalar
        constraint (start) index."""

        self._mosekQuadConIndex = dict()
        """Maps a PICOS quadratic constraint to a MOSEK constraint index."""

        self._mosekCone = dict()
        """
        Maps a PICOS SOCC or RSOCC to a triple:

        - The first entry is the index of a MOSEK conic constraint.
        - The second entry is a list of MOSEK scalar (auxiliary) variable
          indices that appear in the MOSEK conic constraint.
        - The third entry is another list of same size containing auxiliary
          constraints (or None). If an entry in the second list is None, then
          the corresponding index in the first list is of a proper MOSEK scalar
          variable instead of an auxiliary one, which can happen at most once
          per variable and only if the structure of the PICOS constraint allows
          it (that is, if the repspective entry of the cone member happens to be
          a PICOS scalar variable as opposed to a composed affine expression).
        """

        self._mosekLMI = dict()
        """Maps a PICOS LMI to a pair representing its MOSEK representation. The
        first entry is the index of a MOSEK symmetric PSD "bar variable", the
        second entry is the offset for a number of scalar linear equalities."""

        self._mosekBarUnitCoefs = dict()
        """Maps the row (column) count of a symmetric matrix to a list of MOSEK
        symmetric coefficient matrices. More precisely, if X is a n×n symmetric
        matrix, 0 ≤ k ≤ n*(n+1)/2, then <_mosekBarUnitCoefs[n][k], X> = h(X)[k]
        where h refers to the lower-triangular, column-major half-vectorization
        of X. These matrices are used to represent LMIs. Unlike values of
        _mosekLMI, they are shared between LMIs of same size."""

    def reset_problem(self):
        """Implement :meth:`~.solver.Solver.reset_problem`."""
        self.int = None

        self._mosekVarOffset.clear()
        self._mosekLinConOffset.clear()
        self._mosekQuadConIndex.clear()
        self._mosekCone.clear()
        self._mosekLMI.clear()
        self._mosekBarUnitCoefs.clear()

    @classmethod
    def _get_environment(cls):
        if not hasattr(cls, "mosekEnvironment"):
            import mosek
            cls.mosekEnvironment = mosek.Env()

        return cls.mosekEnvironment

    env = property(lambda self: self.__class__._get_environment())
    """This references a MOSEK environment, which is shared among all
    MOSEKSolver instances. (The MOSEK documentation states that "[a]ll tasks in
    the program should share the same environment.")"""

    @classmethod
    def _get_major_version(cls):
        if not hasattr(cls, "mosekVersion"):
            import mosek
            cls.mosekVersion = mosek.Env.getversion()

        return cls.mosekVersion[0]

    ver = property(lambda self: self.__class__._get_major_version())
    """The major version of the available MOSEK library."""

    @staticmethod
    def _streamprinter(text):
        sys.stdout.write(text)
        sys.stdout.flush()

    @staticmethod
    def _low_tri_indices(rowCount):
        """Yield lower triangular (row, col) indices in column-major order."""
        for col in range(rowCount):
            for row in range(col, rowCount):
                yield (row, col)

    def _scalar_affinexp_pic2msk(self, picosExpression):
        assert len(picosExpression) == 1
        return picosExpression.sparse_rows(self._mosekVarOffset)[0]

    def _affinexp_pic2msk(self, picosExpression):
        return picosExpression.sparse_rows(self._mosekVarOffset)

    def _quadexp_pic2msk(self, picosExpression):
        """Transform a quadratic expression from PICOS to MOSEK.

        Tranforms the quadratic part of a PICOS quadratic expression to a
        symmetric, sparse biliniar form of which only the lower triangular
        entries are given, and that can be used with MOSEK's variable vector.
        Note that MOSEK applies a built-in factor of 0.5 to all biliniar forms
        while PICOS doesn't, so a factor of 2 is applied here to cancel it out.
        """
        assert isinstance(picosExpression, QuadraticExpression)
        numVars = self.int.getnumvar()

        # Make a sparse representation of the strict lower, diagonal and strict
        # upper parts of the matrix.
        IL, JL, VL = [], [], []
        ID, JD, VD = [], [], []
        IU, JU, VU = [], [], []
        for (picosVar1, picosVar2), picosCoefs \
        in picosExpression.quadratic_forms.items():
            for sparseIndex in range(len(picosCoefs)):
                localVar1Index   = picosCoefs.I[sparseIndex]
                localVar2Index   = picosCoefs.J[sparseIndex]
                localCoefficient = picosCoefs.V[sparseIndex]
                mskVar1Index = self._mosekVarOffset[picosVar1] + localVar1Index
                mskVar2Index = self._mosekVarOffset[picosVar2] + localVar2Index

                if   mskVar2Index < mskVar1Index:
                    I, J, V = IL, JL, VL
                elif mskVar1Index < mskVar2Index:
                    I, J, V = IU, JU, VU
                else:
                    I, J, V = ID, JD, VD

                I.append(mskVar1Index)
                J.append(mskVar2Index)
                V.append(localCoefficient)

        # Compute the lower triangular part of the biliniar form.
        L = cvxopt.spmatrix(VL, IL, JL, (numVars, numVars))
        D = cvxopt.spmatrix(VD, ID, JD, (numVars, numVars))
        U = cvxopt.spmatrix(VU, IU, JU, (numVars, numVars))
        Q = 2*D + L + U.T

        # Return it as a sparse triple for MOSEK to consume.
        return list(Q.I), list(Q.J), list(Q.V)

    def _import_variable(self, picosVar):
        import mosek

        numVars = self._mosekVarOffset[picosVar] = self.int.getnumvar()
        dim     = picosVar.dim
        indices = range(numVars, numVars + dim)
        self.int.appendvars(dim)

        # Set the variable type.
        if isinstance(picosVar, (IntegerVariable, BinaryVariable)):
            self.int.putvartypelist(
                indices, [mosek.variabletype.type_int]*dim)

        # Import bounds, including the implied bounds of a BinaryVariable.
        if isinstance(picosVar, BinaryVariable):
            self.int.putvarboundlist(
                indices, [mosek.boundkey.ra]*dim, [0]*dim, [1]*dim)
        else:
            boundKeys   = [mosek.boundkey.fr]*dim
            lowerBounds = [0.0]*dim
            upperBounds = [0.0]*dim

            lower, upper = picosVar.bound_dicts

            for i, b in lower.items():
                boundKeys[i] = mosek.boundkey.lo
                lowerBounds[i] = b

            for i, b in upper.items():
                if boundKeys[i] == mosek.boundkey.fr:
                    boundKeys[i] = mosek.boundkey.up
                else:  # Also has a lower bound.
                    if lowerBounds[i] == b:
                        boundKeys[i] = mosek.boundkey.fx
                    else:
                        boundKeys[i] = mosek.boundkey.ra

                upperBounds[i] = b

            self.int.putvarboundlist(
                indices, boundKeys, lowerBounds, upperBounds)

    def _import_linear_constraint(self, picosConstraint):
        import mosek

        numCons = self.int.getnumcon()
        conLen  = len(picosConstraint)
        self.int.appendcons(conLen)

        rows = picosConstraint.sparse_Ab_rows(self._mosekVarOffset)

        if picosConstraint.is_equality():
            boundKey = mosek.boundkey.fx
        elif picosConstraint.is_increasing():
            boundKey = mosek.boundkey.up
        else:
            boundKey = mosek.boundkey.lo

        for localConIndex, (mosekVarIndices, coefs, offset) in enumerate(rows):
            mosekConIndex = numCons + localConIndex

            self.int.putarow(mosekConIndex, mosekVarIndices, coefs)
            self.int.putconbound(mosekConIndex, boundKey, offset, offset)

        self._mosekLinConOffset[picosConstraint] = numCons

    def _import_quad_constraint(self, picosConstraint):
        # Import the linear part first.
        picosLinConPart = picosConstraint.le0.aff < 0
        self._import_linear_constraint(picosLinConPart)
        mosekConIndex = self._mosekLinConOffset.pop(picosLinConPart)

        # Add the quadratic part.
        self.int.putqconk(
            mosekConIndex, *self._quadexp_pic2msk(picosConstraint.le0))

        self._mosekQuadConIndex[picosConstraint] = mosekConIndex

    def _var_was_used_in_cone(self, mosekVariableIndex, usedJustNow=[]):
        if mosekVariableIndex in usedJustNow:
            return True
        for _, mosekVarIndices, _ in self._mosekCone.values():
            if mosekVariableIndex in mosekVarIndices:
                return True
        return False

    def _import_quad_conic_constraint(self, picosConstraint):
        import mosek

        isRotated = isinstance(picosConstraint, RSOCConstraint)
        mosekVars, mosekCons = [], [None]*len(picosConstraint)

        # Get an initial MOSEK representation of the cone member.
        entries = []
        if isRotated:
            # MOSEK internally adds a factor of 2 to the upper bound while PICOS
            # doesn't, so cancel it out by adding a factor of 1/sqrt(2) to both
            # factors of the upper bound.
            f = 1.0 / math.sqrt(2.0)
            entries.append(self._scalar_affinexp_pic2msk(f*picosConstraint.ub1))
            entries.append(self._scalar_affinexp_pic2msk(f*picosConstraint.ub2))
        else:
            entries.append(self._scalar_affinexp_pic2msk(picosConstraint.ub))
        entries += self._affinexp_pic2msk(picosConstraint.ne)

        # Map cone member entries to existing MOSEK variables, if possible.
        mosekVarsMissing = []
        for scalarVarNum, (indices, values, offset) in enumerate(entries):
            if len(values) == 1 and values[0] == 1.0 and offset == 0.0 \
            and not self._var_was_used_in_cone(indices[0], mosekVars):
                mosekVars.append(indices[0])
            else:
                mosekVars.append(None)
                mosekVarsMissing.append(scalarVarNum)

        # Create auxiliary variables and constraints.
        numAux = len(mosekVarsMissing)
        auxVarOffset = self.int.getnumvar()
        auxConOffset = self.int.getnumcon()
        self.int.appendvars(numAux)
        self.int.appendcons(numAux)

        # Mosek fixes (!) new variables at zero, so set them free.
        self.int.putvarboundlist(range(auxVarOffset, auxVarOffset + numAux),
            [mosek.boundkey.fr]*numAux, [0.0]*numAux, [0.0]*numAux)

        # Constrain the auxiliary variables to be equal to the cone member
        # entries for which no existing MOSEK variable could be used.
        for auxNum, missingVarIndex in enumerate(mosekVarsMissing):
            auxVarIndex = auxVarOffset + auxNum
            auxConIndex = auxConOffset + auxNum

            # Prepare the auxiliary constraint.
            indices, values, offset = entries[missingVarIndex]

            # TODO: Instead of always creating a constraint, fix variables via
            #       their bound whenever possible.
            # if len(indices) is 0:
            #     if self._debug():
            #         self._debug("  Fixing MOSEK auxiliary variable: "
            #                 "x[{}] = {}".format(auxVarIndex, offset))
            #
            #     self.int.putvarbound(
            #         auxVarIndex, mosek.boundkey.fx, offset, offset)
            # else:
            indices.append(auxVarIndex)
            values.append(-1.0)

            if self._debug():
                self._debug("  Adding MOSEK auxiliary constraint: "
                    "{}.T * x{} = {}".format(values, indices, -offset))

            # Add the auxiliary constraint.
            self.int.putarow(auxConIndex, indices, values)
            self.int.putconbound(
                auxConIndex, mosek.boundkey.fx, -offset, -offset)

            # Complete the mapping of cone member entries to MOSEK (auxiliary)
            # variables (and auxiliary constraints).
            mosekVars[missingVarIndex] = auxVarIndex
            mosekCons[missingVarIndex] = auxConIndex

        if self._debug():
            self._debug("  Adding MOSEK conic constraint: {} in {}".format(
                mosekVars, "Qr" if isRotated else "Q"))

        # Add the conic constraint.
        coneIndex = self.int.getnumcone()
        mosekCone = mosek.conetype.rquad if isRotated else mosek.conetype.quad
        self.int.appendcone(mosekCone, 0.0, mosekVars)

        self._mosekCone[picosConstraint] = (coneIndex, mosekVars, mosekCons)

    def _import_sdp_constraint(self, picosConstraint):
        import mosek

        n = picosConstraint.size[0]
        dim = (n*(n + 1)) // 2

        # MOSEK does not support general LMIs but so called "bar vars" which
        # are variables in the symmetric positive semidefinite cone. We use them
        # in combination with linear equalities to represent the LMI.
        barVar = self.int.getnumbarvar()
        mosekConOffset = self.int.getnumcon()
        self.int.appendbarvars([n])
        self.int.appendcons(dim)

        # MOSEK uses a storage of symmetric coefficient matrices that are used
        # as dot product coefficients to build scalar constraints involving both
        # "bar vars" and normal scalar variables. We build a couple of these
        # matrices to be able to select individual entries of our "bar vars".
        # More precisely, if X is a n×n symmetric matrix and 0 ≤ k ≤ n*(n+1)/2,
        # then <Units[n][k],X> = h(X)[k] where h refers to the lower-triangular,
        # column-major half-vectorization of X.
        if n in self._mosekBarUnitCoefs:
            Units = self._mosekBarUnitCoefs[n]
        else:
            Units = self._mosekBarUnitCoefs[n] = [
                self.int.appendsparsesymmat(
                    n, [row], [col], [1.0 if row == col else 0.5])
                for row, col in self._low_tri_indices(n)]

        # We iterate over the lower triangular scalar sub-expressions of the
        # expression that the PICOS constraint states to be PSD, and constrain
        # them to be eqal to the MOSEK "bar var" at the same index.
        psd = picosConstraint.psd
        psdRows = psd.sparse_rows(self._mosekVarOffset, lowerTriangle=True)
        for lowTriIndex, (row, col) in enumerate(self._low_tri_indices(n)):
            mosekConIndex = mosekConOffset + lowTriIndex
            indices, coefficients, offset = psdRows[lowTriIndex]

            # The lower-triangular entries in the PSD-constrained matrix …
            self.int.putarow(mosekConIndex, indices, coefficients)

            # … minus the corresponding bar var entries …
            self.int.putbaraij(
                mosekConIndex, barVar, [Units[lowTriIndex]], [-1.0])

            # … should equal zero.
            self.int.putconbound(
                mosekConIndex, mosek.boundkey.fx, -offset, -offset)

            if self._debug():
                self._debug("  Index {} ({}, {}): indices = {}, coefs = {}"
                    .format(lowTriIndex, row, col, indices, coefficients))

        self._mosekLMI[picosConstraint] = (barVar, mosekConOffset)

    def _import_constraint(self, picosConstraint):
        if self._debug():
            self._debug("Importing Constraint: {}".format(picosConstraint))

        if   isinstance(picosConstraint, AffineConstraint):
            self._import_linear_constraint(picosConstraint)
        elif isinstance(picosConstraint, ConvexQuadraticConstraint):
            self._import_quad_constraint(picosConstraint)
        elif isinstance(picosConstraint, SOCConstraint) \
        or   isinstance(picosConstraint, RSOCConstraint):
            self._import_quad_conic_constraint(picosConstraint)
        elif isinstance(picosConstraint, LMIConstraint):
            self._import_sdp_constraint(picosConstraint)
        else:
            assert False, "Constraint type not supported."

    def _reset_objective(self):
        # Reset affine part.
        numVars = self.int.getnumvar()
        self.int.putclist(range(numVars), [0.0]*numVars)
        self.int.putcfix(0.0)

        # Reset quadratic part.
        self.int.putqobj([], [], [])

    def _import_affine_objective(self, picosObjective):
        mosekIndices      = []
        mosekCoefficients = []

        for picosVar, picosCoef in picosObjective._coefs.items():
            for localIndex in range(picosVar.dim):
                if picosCoef[localIndex]:
                    mosekIndex = self._mosekVarOffset[picosVar] + localIndex
                    mosekIndices.append(mosekIndex)
                    mosekCoefficients.append(picosCoef[localIndex])

        self.int.putclist(mosekIndices, mosekCoefficients)

        if picosObjective._const:
            self.int.putcfix(picosObjective._const[0])

    def _import_quadratic_objective(self, picosObjective):
        # Import the quadratic part.
        self.int.putqobj(*self._quadexp_pic2msk(picosObjective))

        # Import the affine part.
        self._import_affine_objective(picosObjective.aff)

    def _import_objective(self):
        import mosek

        picosSense, picosObjective = self.ext.no

        # Import objective sense.
        if picosSense == "min":
            self.int.putobjsense(mosek.objsense.minimize)
        else:
            assert picosSense == "max"
            self.int.putobjsense(mosek.objsense.maximize)

        # Import objective function.
        if isinstance(picosObjective, AffineExpression):
            self._import_affine_objective(picosObjective)
        else:
            assert isinstance(picosObjective, QuadraticExpression)
            self._import_quadratic_objective(picosObjective)

    def _import_problem(self):
        # Create a problem instance.
        self.int = self.env.Task()

        # Import variables.
        for variable in self.ext.variables.values():
            self._import_variable(variable)

        # Import constraints.
        for constraint in self.ext.constraints.values():
            self._import_constraint(constraint)

        # Set objective.
        self._import_objective()

    def _update_problem(self):
        for oldConstraint in self._removed_constraints():
            raise ProblemUpdateError("PICOS does not support removing "
                "constraints from a MOSEK instance.")

        for oldVariable in self._removed_variables():
            raise ProblemUpdateError("PICOS does not support removing variables"
                " from a MOSEK instance.")

        for newVariable in self._new_variables():
            self._import_variable(newVariable)

        for newConstraint in self._new_constraints():
            self._import_constraint(newConstraint)

        if self._objective_has_changed():
            self._reset_objective()
            self._import_objective()

    def _solve(self):
        import mosek

        # Determine whether a basic solution will be available.
        # TODO: Give Problem an interface for checks like this.
        isLP = isinstance(self.ext.no.function, AffineExpression) \
            and all(isinstance(constraint, AffineConstraint)
            for constraint in self.ext.constraints.values())

        # Reset all solver options to default.
        self.int.setdefaults()

        # verbosity
        if self._verbose():
            self.int.set_Stream(mosek.streamtype.log, self._streamprinter)
            self.int.putintparam(mosek.iparam.log, self.ext.verbosity())

        # abs_prim_fsb_tol
        if self.ext.options.abs_prim_fsb_tol is not None:
            value = self.ext.options.abs_prim_fsb_tol

            # Interior-point primal feasibility tolerances.
            self.int.putdouparam(mosek.dparam.intpnt_tol_pfeas,    value)
            self.int.putdouparam(mosek.dparam.intpnt_co_tol_pfeas, value)
            self.int.putdouparam(mosek.dparam.intpnt_qo_tol_pfeas, value)
            if self.ver <= 8:
                self.int.putdouparam(mosek.dparam.intpnt_nl_tol_pfeas, value)

            # Simplex primal feasibility tolerance.
            self.int.putdouparam(mosek.dparam.basis_tol_x, value)

            # Mixed-integer (primal) feasibility tolerance.
            self.int.putdouparam(mosek.dparam.mio_tol_feas, value)

        # abs_dual_fsb_tol
        if self.ext.options.abs_dual_fsb_tol is not None:
            value = self.ext.options.abs_dual_fsb_tol

            # Interior-point dual feasibility tolerances.
            self.int.putdouparam(mosek.dparam.intpnt_tol_dfeas,    value)
            self.int.putdouparam(mosek.dparam.intpnt_co_tol_dfeas, value)
            self.int.putdouparam(mosek.dparam.intpnt_qo_tol_dfeas, value)
            if self.ver <= 8:
                self.int.putdouparam(mosek.dparam.intpnt_nl_tol_dfeas, value)

            # Simplex dual feasibility (optimality) tolerance.
            self.int.putdouparam(mosek.dparam.basis_tol_s, value)

        # rel_dual_fsb_tol
        if self.ext.options.rel_dual_fsb_tol is not None:
            # Simplex relative dual feasibility (optimality) tolerance.
            self.int.putdouparam(mosek.dparam.basis_rel_tol_s,
                self.ext.options.rel_dual_fsb_tol)

        # rel_ipm_opt_tol
        if self.ext.options.rel_ipm_opt_tol is not None:
            value = self.ext.options.rel_ipm_opt_tol

            # Interior-point primal feasibility tolerances.
            self.int.putdouparam(mosek.dparam.intpnt_tol_rel_gap,    value)
            self.int.putdouparam(mosek.dparam.intpnt_co_tol_rel_gap, value)
            self.int.putdouparam(mosek.dparam.intpnt_qo_tol_rel_gap, value)
            if self.ver <= 8:
                self.int.putdouparam(mosek.dparam.intpnt_nl_tol_rel_gap, value)

        # abs_bnb_opt_tol
        if self.ext.options.abs_bnb_opt_tol is not None:
            self.int.putdouparam(mosek.dparam.mio_tol_abs_gap,
                self.ext.options.abs_bnb_opt_tol)

        # rel_bnb_opt_tol
        if self.ext.options.rel_bnb_opt_tol is not None:
            self.int.putdouparam(mosek.dparam.mio_tol_rel_gap,
                self.ext.options.rel_bnb_opt_tol)

        # integrality_tol
        if self.ext.options.integrality_tol is not None:
            self.int.putdouparam(mosek.dparam.mio_tol_abs_relax_int,
                self.ext.options.integrality_tol)

        # max_iterations
        if self.ext.options.max_iterations is not None:
            value = self.ext.options.max_iterations
            self.int.putintparam(mosek.iparam.bi_max_iterations,     value)
            self.int.putintparam(mosek.iparam.intpnt_max_iterations, value)
            self.int.putintparam(mosek.iparam.sim_max_iterations,    value)

        _lpm = {
            "interior": (mosek.optimizertype.intpnt if isLP
                else mosek.optimizertype.conic),
            "psimplex": mosek.optimizertype.primal_simplex,
            "dsimplex": mosek.optimizertype.dual_simplex}

        # lp_node_method
        if self.ext.options.lp_node_method is not None:
            value = self.ext.options.lp_node_method
            assert value in _lpm, "Unexpected lp_node_method value."
            self.int.putintparam(mosek.iparam.mio_node_optimizer, _lpm[value])

        # lp_root_method
        if self.ext.options.lp_root_method is not None:
            value = self.ext.options.lp_root_method
            assert value in _lpm, "Unexpected lp_root_method value."
            self.int.putintparam(mosek.iparam.mio_root_optimizer, _lpm[value])

        # timelimit
        if self.ext.options.timelimit is not None:
            value = float(self.ext.options.timelimit)
            self.int.putdouparam(mosek.dparam.optimizer_max_time, value)
            self.int.putdouparam(mosek.dparam.mio_max_time,       value)

        # max_fsb_nodes
        if self.ext.options.max_fsb_nodes is not None:
            self.int.putintparam(mosek.iparam.mio_max_num_solutions,
                self.ext.options.max_fsb_nodes)

        # Handle MOSEK-specific options.
        for option, value in self.ext.options.mosek_params.items():
            try:
                self.int.putparam(option.upper(), str(value))
            except mosek.Error as error:
                self._handle_bad_option_value("mosek_params", str(error))

        # Handle unsupported options.
        # TODO: Handle "hotstart" option (via mio_construct_sol).
        self._handle_unsupported_option("hotstart", "treememory")

        # Attempt to solve the problem.
        with self._header(), self._stopwatch():
            try:
                self.int.optimize()
            except mosek.Error as error:
                if error.errno in (
                        mosek.rescode.err_con_q_not_psd,
                        mosek.rescode.err_con_q_not_nsd,
                        mosek.rescode.err_obj_q_not_psd,
                        mosek.rescode.err_obj_q_not_nsd,
                        mosek.rescode.err_toconic_constr_q_not_psd,
                        mosek.rescode.err_toconic_objective_not_psd):
                    self._handle_continuous_nonconvex_error(error)
                else:
                    raise

        # Set the solution to be retrieved.
        if self.ext.is_continuous():
            if isLP:
                solType = mosek.soltype.bas
            else:
                solType = mosek.soltype.itr
        else:
            solType = mosek.soltype.itg

        # Retrieve primals.
        primals = {}
        if self.ext.options.primals is not False:
            values = [float("nan")]*self.int.getnumvar()
            self.int.getxx(solType, values)

            for picosVar in self.ext.variables.values():
                mosekOffset = self._mosekVarOffset[picosVar]
                primal = values[mosekOffset:mosekOffset + picosVar.dim]

                if float("nan") in primal:
                    primals[picosVar] = None
                else:
                    primals[picosVar] = primal

        # Retrieve duals.
        duals = {}
        if self.ext.options.duals is not False and self.ext.is_continuous():
            minimizing = self.int.getobjsense() == mosek.objsense.minimize

            for constraint in self.ext.constraints.values():
                length = len(constraint)
                dual   = [float("nan")]*length
                if isinstance(constraint, AffineConstraint):
                    offset = self._mosekLinConOffset[constraint]
                    self.int.getyslice(solType, offset, offset + length, dual)
                elif isinstance(constraint, ConvexQuadraticConstraint):
                    # TODO: Implement consistent QCQP dual retrieval for all
                    #       solvers that return such duals.
                    dual = None
                elif isinstance(constraint, SOCConstraint) \
                or isinstance(constraint, RSOCConstraint):
                    mosekVars = self._mosekCone[constraint][1]
                    for localConeIndex in range(length):
                        x = [float("nan")]
                        offset = mosekVars[localConeIndex]
                        self.int.getsnxslice(solType, offset, offset + 1, x)
                        dual[localConeIndex] = x[0]

                    if isinstance(constraint, SOCConstraint):
                        dual = [-du for du in dual]
                    elif isinstance(constraint, RSOCConstraint):
                        dual[0] = -dual[0] / math.sqrt(2.0)
                        dual[1] = -dual[1] / math.sqrt(2.0)
                        dual[2:] = [-du for du in dual[2:]]
                elif isinstance(constraint, LMIConstraint):
                    n = constraint.size[0]
                    barVar, _ = self._mosekLMI[constraint]
                    lowerTriangularDual = [float("nan")]*(n*(n + 1) // 2)
                    self.int.getbarsj(solType, barVar, lowerTriangularDual)
                    for lti, (row, col) in enumerate(self._low_tri_indices(n)):
                        value = lowerTriangularDual[lti]
                        dual[n*row + col] = value
                        if row != col:
                            dual[n*col + row] = value
                else:
                    assert False, "Constraint type not supported."

                if dual is None:
                    pass
                elif float("nan") in dual:
                    dual = None
                else:
                    dual = cvxopt.matrix(dual, constraint.size)

                    if isinstance(constraint, AffineConstraint) \
                    or isinstance(constraint, LMIConstraint):
                        if constraint.is_decreasing():
                            dual = -dual

                    if minimizing:
                        dual = -dual

                duals[constraint] = dual

        primalStatus  = self._solution_status_pic2msk(
            self.int.getsolsta(solType), primalSolution=True)
        dualStatus    = self._solution_status_pic2msk(
            self.int.getsolsta(solType), primalSolution=False)
        problemStatus = self._problem_status_pic2msk(
            self.int.getprosta(solType), not self.ext.is_continuous())

        return self._make_solution(
            primals, duals, primalStatus, dualStatus, problemStatus)

    def _solution_status_pic2msk(self, statusCode, primalSolution):
        from mosek import solsta as ss
        dualSolution = not primalSolution

        map = {
            ss.unknown:                 SS_UNKNOWN,
            ss.optimal:                 SS_OPTIMAL,
            ss.prim_feas:
                SS_FEASIBLE   if primalSolution else SS_UNKNOWN,
            ss.dual_feas:
                SS_FEASIBLE   if dualSolution   else SS_UNKNOWN,
            ss.prim_and_dual_feas:      SS_FEASIBLE,
            ss.prim_infeas_cer:
                SS_INFEASIBLE if primalSolution else SS_UNKNOWN,
            ss.dual_infeas_cer:
                SS_INFEASIBLE if dualSolution   else SS_UNKNOWN,
            ss.prim_illposed_cer:       SS_UNKNOWN,
            ss.dual_illposed_cer:       SS_UNKNOWN,
            ss.integer_optimal:         SS_OPTIMAL,
        }

        if self.ver < 9:
            map.update({
                ss.near_optimal:            SS_FEASIBLE,
                ss.near_prim_feas:          SS_UNKNOWN,
                ss.near_dual_feas:          SS_UNKNOWN,
                ss.near_prim_and_dual_feas: SS_UNKNOWN,
                ss.near_prim_infeas_cer:    SS_UNKNOWN,
                ss.near_dual_infeas_cer:    SS_UNKNOWN,
                ss.near_integer_optimal:    SS_FEASIBLE
            })

        try:
            return map[statusCode]
        except KeyError:
            self._warn(
                "The MOSEK solution status code {} is not known to PICOS."
                .format(statusCode))
            return SS_UNKNOWN

    def _problem_status_pic2msk(self, statusCode, integerProblem):
        from mosek import prosta as ps

        map = {
            ps.unknown:                  PS_UNKNOWN,
            ps.prim_and_dual_feas:       PS_FEASIBLE,
            ps.prim_feas:
                PS_FEASIBLE if integerProblem else PS_UNKNOWN,
            ps.dual_feas:                PS_UNKNOWN,
            ps.prim_infeas:              PS_INFEASIBLE,
            ps.dual_infeas:              PS_UNBOUNDED,  # TODO: UNB_OR_INF
            ps.prim_and_dual_infeas:     PS_INFEASIBLE,
            ps.ill_posed:                PS_ILLPOSED,
            ps.prim_infeas_or_unbounded: PS_UNKNOWN  # TODO: UNB_OR_INF
        }

        if self.ver < 9:
            map.update({
                ps.near_prim_and_dual_feas:  PS_UNKNOWN,
                ps.near_prim_feas:           PS_UNKNOWN,
                ps.near_dual_feas:           PS_UNKNOWN
            })

        try:
            return map[statusCode]
        except KeyError:
            self._warn("The MOSEK problem status code {} is not known to PICOS."
                .format(statusCode))
            return PS_UNKNOWN


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`FlowConstraint`."""

from collections import namedtuple

from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class FlowConstraint(Constraint):
    """Network flow constraint.

    .. note ::
        Unlike other :class:`~.constraint.Constraint` implementations, this one
        is instanciated by the user (via a wrapper function), so it is raising
        exceptions instead of making assertions.
    """

    class Conversion(ConstraintConversion):
        """Network flow constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import RealVariable
            from . import AffineConstraint

            numNodes, numEdges, flowCons, hasCapacities = subtype
            V, E = numNodes, numEdges

            # Capacity and non-negativity constraints.
            yield ("con", AffineConstraint.make_type(dim=1, eq=False),
                2*E if hasCapacities else E)

            if len(flowCons) == 1:
                yield ("con", AffineConstraint.make_type(dim=1, eq=True), V - 1)
            else:
                for k in flowCons:
                    yield ("var", RealVariable.make_var_type(dim=1, bnd=0), E)

                    for addition in cls.predict((V, E, (k,), False), options):
                        yield addition

                yield ("con", AffineConstraint.make_type(dim=1, eq=True), E)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..modeling import Problem
            P = Problem()
            return cls._convert(con, P)  # Allow for recursion.

        @classmethod
        def _convert(cls, con, P):
            from ..expressions import new_param, sum

            G          = con.graph
            f          = con.f
            source     = con.source
            sink       = con.sink
            flow_value = con.flow_value
            capacity   = con.capacity
            graphName  = con.graphName

            # Add capacity constraints.
            if capacity is not None:
                c = {}
                for v, w, data in G.edges(data=True):
                    c[(v, w)] = data[capacity]
                c = new_param('c', c)

                P.add_list_of_constraints([f[e] <= c[e] for e in G.edges()])

            # Add non-negativity constraints.
            P.add_list_of_constraints([f[e] >= 0 for e in G.edges()])

            # One source, one sink.
            if not isinstance(source, list) and not isinstance(sink, list):
                # Add flow conversation constrains.
                P.add_list_of_constraints([
                    sum([f[p, i] for p in G.predecessors(i)])
                    == sum([f[i, j] for j in G.successors(i)])
                    for i in G.nodes() if i != sink and i != source])

                # Source flow at S
                P.add_constraint(
                    sum([f[p, source] for p in G.predecessors(source)])
                    + flow_value ==
                    sum([f[source, j] for j in G.successors(source)]))

            # One source, multiple sinks.
            elif not isinstance(source, list):
                # Add flow conversation constrains.
                P.add_list_of_constraints([
                    sum([f[p, i] for p in G.predecessors(i)])
                    == sum([f[i, j] for j in G.successors(i)])
                    for i in G.nodes() if i not in sink and i != source])

                for k in range(0, len(sink)):
                    # Sink flow at T
                    P.add_constraint(
                        sum([f[p, sink[k]] for p in G.predecessors(sink[k])])
                        == sum([f[sink[k], j] for j in G.successors(sink[k])])
                        + flow_value[k])

            # Multiple sources, one sink.
            elif not isinstance(sink, list):
                # Add flow conversation constrains.
                P.add_list_of_constraints([
                    sum([f[p, i] for p in G.predecessors(i)])
                    == sum([f[i, j] for j in G.successors(i)])
                    for i in G.nodes() if i not in source and i != sink])

                for k in range(0, len(source)):
                    # Source flow at T
                    P.add_constraint(sum(
                        [f[p, source[k]] for p in G.predecessors(source[k])])
                        + flow_value[k] ==
                        sum([f[source[k], j] for j in G.successors(source[k])]))

            # Multiple sources, multiple sinks.
            # TODO: Recursion adds redundant non-negativity constraints.
            elif isinstance(sink, list) and isinstance(source, list):
                SS = list(set(source))
                TT = list(set(sink))

                if len(SS) <= len(TT):
                    ftmp = {}
                    for s in SS:
                        ftmp[s] = {}
                        sinks_from_s = [
                            t for (i, t) in enumerate(sink) if source[i] == s]
                        values_from_s = [v for (i, v)
                            in enumerate(flow_value) if source[i] == s]

                        for e in G.edges():
                            ftmp[s][e] = P.add_variable(
                                '__f[{0}][{1}]'.format(s, e), 1)

                        # Immediately convert another FlowConstraint so that the
                        # reformulation created from this conversion doesn't
                        # need to be run twice in a row.
                        cls._convert(cls(
                            G, ftmp[s], source=s, sink=sinks_from_s,
                            flow_value=values_from_s, graphName=graphName), P)

                    P.add_list_of_constraints([
                        f[e] == sum([ftmp[s][e] for s in SS])
                        for e in G.edges()])
                else:
                    ftmp = {}
                    for t in TT:
                        ftmp[t] = {}
                        sources_to_t = [
                            s for (i, s) in enumerate(source) if sink[i] == t]
                        values_to_t = [v for (i, v) in enumerate(flow_value)
                            if sink[i] == t]

                        for e in G.edges():
                            ftmp[t][e] = P.add_variable(
                                '__f[{0}][{1}]'.format(t, e), 1)

                        # Immediately convert another FlowConstraint so that the
                        # reformulation created from this conversion doesn't
                        # need to be run twice in a row.
                        cls._convert(cls(
                            G, ftmp[t], source=sources_to_t, sink=t,
                            flow_value=values_to_t, graphName=graphName), P)

                    P.add_list_of_constraints([
                        f[e] == sum([ftmp[t][e] for t in TT])
                        for e in G.edges()])

            else:
                assert False, "Dijkstra-IF fallthrough."

            return P

    def __init__(
            self, G, f, source, sink, flow_value, capacity=None, graphName=""):
        """Construct a network flow constraint.

        :param G: A directed graph.
        :type G: `networkx DiGraph <http://networkx.lanl.gov/index.html>`_.

        :param dict f: A dictionary of variables indexed by the edges of ``G``.

        :param source: Either a node of ``G`` or a list of nodes in case of a
            multi-source flow.

        :param sink: Either a node of ``G`` or a list of nodes in case of a
            multi-sink flow.

        :param flow_value: The value of the flow, or a list of values in case of
            a single-source/multi-sink flow. In the latter case, the values
            represent the demands of each sink (resp. of each source for a
            multi-source/single-sink flow). The values can be either constants
            or :class:`~picos.expressions.AffineExpression`.

        :param capacity: Either ``None`` or a string. If this is a string, it
            indicates the key of the edge dictionaries of ``G`` that is used for
            the capacity of the links. Otherwise, edges have an unbounded
            capacity.

        :param str graphName: Name of the graph as used in the string
            representation of the constraint.
        """
        if len(f) != len(G.edges()):
            raise ValueError(
                "The number of variables does not match the number of edges.")

        if isinstance(sink, list) and len(sink) == 1:
            source = source[0]

        if isinstance(sink, list) and len(sink) == 1:
            sink = sink[0]

        if isinstance(source, list) and len(source) != len(flow_value):
            raise ValueError("The number of sources does not match the number "
                "of flow values.")

        if isinstance(sink, list) and len(sink) != len(flow_value):
            raise ValueError("The number of sinks does not match the number "
                "of flow values.")

        if isinstance(source, list) and isinstance(sink, list) \
        and len(sink) != len(source):
            raise ValueError("The number of sinks does not match the number "
                "of sources.")

        self.graph      = G
        self.f          = f
        self.source     = source
        self.sink       = sink
        self.flow_value = flow_value
        self.capacity   = capacity
        self.graphName  = graphName

        # Build the string description.
        if isinstance(source, list):
            sourceStr = "(" + ",".join(source) + ")"
        else:
            sourceStr = str(source)

        if isinstance(sink, list):
            sinkStr = "(" + ",".join(sink) + ")"
        else:
            sinkStr = str(sink)

        if isinstance(flow_value, list):
            valueStr = "values " + ", ".join([v.string if hasattr(v, "string")
                else str(v) for v in flow_value])
        else:
            valueStr = "value " + flow_value.string \
                if hasattr(flow_value, "string") else str(flow_value)

        self.comment = "{}{}-{}-flow{} has {}.".format(
            "Feasible " if capacity is not None else "", sourceStr, sinkStr,
            " in {}".format(graphName) if graphName else "", valueStr)

        super(FlowConstraint, self).__init__("Flow")

    Subtype = namedtuple("Subtype",
        ("lenV", "lenE", "flowCons", "hasCapacities"))

    @classmethod
    def _cost(cls, subtype):
        return subtype.lenE  # Somewhat arbitrary.

    def _subtype(self):
        s = len(self.source) if isinstance(self.source, list) else 1
        t = len(self.sink)   if isinstance(self.sink, list)   else 1
        V = len(self.graph.nodes())
        E = len(self.graph.edges())

        if s == 1 or t == 1:
            flowCons = (max(s, t),)
        else:
            flowCons = []
            S, T = self.source, self.sink
            A, B = set(S), set(T)
            if len(A) > len(B):
                S, T = T, S
                A, B = B, A
            for s in A:
                flowCons.append(S.count(s))
            flowCons = tuple(flowCons)

        assert sum(flowCons) == max(s, t)

        return self.Subtype(V, E, flowCons, self.capacity is not None)

    def _expression_names(self):
        return
        yield

    # HACK: These variables are not stored in named expressions but in a
    #       dictionary. To make prediction work, they must be considered part
    #       of the problem that the flow constraint was added to.
    @property
    def variables(self):  # noqa
        return frozenset(self.f.values())

    # HACK: See above.
    def replace_variables(self, var_map):  # noqa
        f = {key: var_map[var] for key, var in self.f.items()}
        return FlowConstraint(self.graph, f, self.source, self.sink,
            self.flow_value, self.capacity, self.graphName)

    def _str(self):
        return self.comment

    def _get_slack(self):
        raise NotImplementedError

    def draw(self):
        """Draw the graph."""
        G = self.graph

        import networkx as nx
        import matplotlib.pyplot as plt

        pos = nx.spring_layout(G)
        edge_labels = dict([((u, v,), d["capacity"])
                            for u, v, d in G.edges(data=True)])
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)
        nx.draw(G, pos)
        plt.show()


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2020 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`Dualization`."""

from itertools import chain

import cvxopt

from ..apidoc import api_end, api_start
from ..constraints import (AffineConstraint, LMIConstraint, RSOCConstraint,
                           SOCConstraint)
from ..expressions import AffineExpression, Constant
from ..expressions.algebra import rsoc, soc
from ..expressions.variables import (CONTINUOUS_VARTYPES, RealVariable,
                                     SymmetricVariable)
from ..modeling import Problem
from ..modeling.footprint import Footprint, Specification
from ..modeling.solution import PS_INFEASIBLE, PS_UNBOUNDED, Solution
from .reformulation import Reformulation

_API_START = api_start(globals())
# -------------------------------


class Dualization(Reformulation):
    """Lagrange dual problem reformulation."""

    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression],
        variables=CONTINUOUS_VARTYPES,
        constraints=[
            AffineConstraint,
            SOCConstraint,
            RSOCConstraint,
            LMIConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.reformulation.Reformulation.supports`."""
        return footprint.options.dualize and footprint << cls.SUPPORTED

    @classmethod
    def predict(cls, footprint):
        """Implement :meth:`~.reformulation.Reformulation.predict`."""
        vars = []
        cons = []

        # Objective direction.
        dir = "min" if footprint.direction == "max" else "max"

        # Objective function.
        # HACK: Conversion adds a zero-fixed variable to the objective so that
        #       it is never constant. This makes prediction succeed in any case.
        obj = AffineExpression.make_type(
            shape=(1, 1), constant=False, nonneg=False)

        # Variables and their bounds.
        bound_cons = []
        for vartype, k in footprint.variables.items():
            cons.append((AffineConstraint.make_type(
                dim=vartype.subtype.dim, eq=True), k))

            if vartype.subtype.bnd:
                bound_cons.append((AffineConstraint.make_type(
                    dim=vartype.subtype.bnd, eq=False), k))

        # Constraints by type.
        for contype, k in chain(footprint.constraints.items(), bound_cons):
            if contype.clstype is AffineConstraint:
                d = contype.subtype.dim

                # Dual variable.
                vars.append((RealVariable.make_var_type(dim=d, bnd=0), k))

                # Conic membership constraint.
                if not contype.subtype.eq:
                    cons.append((
                        AffineConstraint.make_type(dim=d, eq=False), k))
            elif contype.clstype is SOCConstraint:
                n = contype.subtype.argdim
                d = n + 1

                # Dual variable.
                vars.append((RealVariable.make_var_type(dim=d, bnd=0), k))

                # Conic membership constraint.
                cons.append((SOCConstraint.make_type(argdim=n), k))
            elif contype.clstype is RSOCConstraint:
                n = contype.subtype.argdim
                d = n + 2

                # Dual variable.
                vars.append((RealVariable.make_var_type(dim=d, bnd=0), k))

                # Conic membership constraint.
                cons.append((RSOCConstraint.make_type(argdim=n), k))
            elif contype.clstype is LMIConstraint:
                n = contype.subtype.diag
                d = n*(n + 1)//2

                # Dual variable.
                vars.append((SymmetricVariable.make_var_type(dim=d, bnd=0), k))

                # Conic membership constraint.
                cons.append((LMIConstraint.make_type(diag=n), k))
            else:
                assert False, "Unexpected constraint type."

        # HACK: See above.
        vars.append((RealVariable.make_var_type(dim=1, bnd=2), 1))

        # Option changes.
        nd_opts = footprint.options.updated(
            dualize=False,
            primals=footprint.options.duals,
            duals=footprint.options.primals
        ).nondefaults

        return Footprint.from_types(dir, obj, vars, cons, nd_opts)

    def forward(self):
        """Implement :meth:`~.reformulation.Reformulation.forward`."""
        self._pc2dv = {}  # Maps primal constraints to dual variables.
        self._pv2dc = {}  # Maps primal variables to dual constraints.

        P = self.input

        # Create the dual problem from scratch.
        D = self.output = Problem(copyOptions=P.options)

        # Reset the 'dualize' option so that solvers can load the dual.
        D.options.dualize = False

        # Swap 'primals' and 'duals' option.
        D.options.primals = P.options.duals
        D.options.duals = P.options.primals

        # Retrieve the primal objective in standard form.
        original_primal_direction, primal_objective = P.objective.normalized
        if original_primal_direction == "max":
            primal_objective = -primal_objective

        # Start building the dual objective function.
        dual_objective = primal_objective.cst

        # Start building the dual linear equalities for each primal variable.
        obj_coefs = primal_objective.lin._coefs
        dual_equality_terms = {var: -Constant(obj_coefs[var].T)
            if var in obj_coefs else AffineExpression.zero(var.dim)
            for var in P.variables.values()}

        # Turn variable bounds into affine inequality constraints.
        # NOTE: If bound-free variables are required by another reformulation or
        #       solver in the future, there should be a reformulation for this.
        bound_cons = []
        for primal_var in P.variables.values():
            bound_constraint = primal_var.bound_constraint
            if bound_constraint:
                bound_cons.append(bound_constraint)

        # Handle the primal constraints.
        # NOTE: Constraint conversion works as follows:
        #       1. Transform constraint into conic standard form:
        #          fx = Ax-b >_K 0.
        #       2. Create a dual variable for the constraint.
        #       3. Constrain the dual variable to be a member of the dual
        #          cone K*.
        #       4. Extend the dual's linear equalities for each primal variable.
        #       5. Extend the dual's objective function.
        # TODO: Consider adding a ConicConstraint abstract subclass of
        #       Constraint with a method conic_form that returns a pair of cone
        #       member (affine expression) and cone. The conic standard form
        #       function f(x) and the dual cone can then be obtained by negation
        #       and through a new Cone.dual method, respectively.
        for primal_con in chain(P.constraints.values(), bound_cons):
            if isinstance(primal_con, AffineConstraint):
                # Transform constraint into conic standard form.
                fx = primal_con.ge0.vec

                # Create a dual variable for the constraint.
                self._pc2dv[primal_con] = dual_var = RealVariable(
                    "aff_{}".format(primal_con.id), len(fx))

                # Constrain the dual variable to be a member of the dual cone.
                # NOTE: In the equality case, the dual cone is the real field.
                if primal_con.is_inequality():
                    D.add_constraint(dual_var >= 0)

                # Extend the dual's linear equalities for each primal variable.
                for var, coef in fx._coefs.items():
                    dual_equality_terms[var] += coef.T*dual_var

                # Extend the dual's objective function.
                dual_objective -= fx._const.T*dual_var

            elif isinstance(primal_con, SOCConstraint):
                # Transform constraint into conic standard form.
                fx = (primal_con.ub // primal_con.ne.vec)

                # Create a dual variable for the constraint.
                self._pc2dv[primal_con] = dual_var = RealVariable(
                    "soc_{}".format(primal_con.id), len(fx))

                # Constrain the dual variable to be a member of the dual cone.
                D.add_constraint(dual_var << soc())

                # Extend the dual's linear equalities for each primal variable.
                for var, coef in fx._coefs.items():
                    dual_equality_terms[var] += coef.T*dual_var

                # Extend the dual's objective function.
                dual_objective -= fx._const.T*dual_var
            elif isinstance(primal_con, RSOCConstraint):
                # Transform constraint into conic standard form.
                fx = (primal_con.ub1 // primal_con.ub2 // primal_con.ne.vec)

                # Create a dual variable for the constraint.
                self._pc2dv[primal_con] = dual_var = RealVariable(
                    "rso_{}".format(primal_con.id), len(fx))

                # Constrain the dual variable to be a member of the dual cone.
                D.add_constraint(dual_var << rsoc(4))

                # Extend the dual's linear equalities for each primal variable.
                for var, coef in fx._coefs.items():
                    dual_equality_terms[var] += coef.T*dual_var

                # Extend the dual's objective function.
                dual_objective -= fx._const.T*dual_var
            elif isinstance(primal_con, LMIConstraint):
                # Transform constraint into conic standard form.
                fx = primal_con.psd

                # Create a dual variable for the constraint.
                self._pc2dv[primal_con] = dual_var = SymmetricVariable(
                    "lmi_{}".format(primal_con.id), fx.shape)

                # Constrain the dual variable to be a member of the dual cone.
                D.add_constraint(dual_var >> 0)

                # Extend the dual's linear equalities for each primal variable.
                for var, coef in fx._coefs.items():
                    dual_equality_terms[var] += coef.T*dual_var.vec

                # Extend the dual's objective function.
                dual_objective -= fx._const.T*dual_var.vec
            else:
                assert False, "Unexpected constraint type."

        # Add the finished linear equalities for each variable.
        for var, term in dual_equality_terms.items():
            self._pv2dc[var] = D.add_constraint(term == 0)

        # Adjust dual optimization direction to obtain the duality property.
        if original_primal_direction == "max":
            dual_direction = "min"
            dual_objective = -dual_objective
        else:
            dual_direction = "max"

        # HACK: Add a zero-fixed variable to the objective so that it is never
        #       constant. This makes prediction succeed in any case.
        dual_objective += RealVariable("zero", 1, lower=0, upper=0)

        # Set the finished dual objective.
        D.objective = dual_direction, dual_objective

    def update(self):
        """Implement :meth:`~.reformulation.Reformulation.update`."""
        # TODO: Implement updates to an existing dualization. This is optional.
        raise NotImplementedError

    def backward(self, solution):
        """Implement :meth:`~.reformulation.Reformulation.backward`."""
        P = self.input

        # Require primal values to be properly vectorized.
        if not solution.vectorizedPrimals:
            raise NotImplementedError(
                "Solving with dualize=True is not supported with solvers that "
                "produce unvectorized primal solutions.")

        # Retrieve primals for the primal problem.
        primals = {}
        for var in P.variables.values():
            con = self._pv2dc[var]
            primal = solution.duals[con] if con in solution.duals else None

            if primal is not None:
                primals[var] = cvxopt.matrix(primal)
            else:
                primals[var] = None

        # Retrieve duals for the primal problem.
        duals = {}
        for con in P.constraints.values():
            var = self._pc2dv[con]
            dual = solution.primals[var] if var in solution.primals else None

            if dual is not None:
                duals[con] = var._vec.devectorize(cvxopt.matrix(dual))
            else:
                duals[con] = None

        # Swap infeasible/unbounded for a problem status.
        if solution.problemStatus == PS_UNBOUNDED:
            problemStatus = PS_INFEASIBLE
        elif solution.problemStatus == PS_INFEASIBLE:
            problemStatus = PS_UNBOUNDED
        else:
            problemStatus = solution.problemStatus

        return Solution(
            primals=primals,
            duals=duals,
            problem=solution.problem,
            solver=solution.solver,
            primalStatus=solution.dualStatus,  # Intentional swap.
            dualStatus=solution.primalStatus,  # Intentional swap.
            problemStatus=problemStatus,
            searchTime=solution.searchTime,
            info=solution.info,
            vectorizedPrimals=True)


# --------------------------------------
__all__ = api_end(_API_START, globals())
